/*
 *  Login.app
 * 
 *  Copyright (c) 1997-1999 Per Liden
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <stdio.h>
#include <unistd.h>
#include "preferences.h"
#include "util.h"
#include "resources.h"
#include "graphics.h"
#include "config.h"


static RImage* loadImage(const char* appPath, const char* theme, const char* file) {    
    RImage* image;
    char*   themeDir;
    char*   imageFile;

    themeDir = THEME_DIR;
    imageFile = alloc(strlen(appPath) + strlen(themeDir) + strlen(theme) + strlen(file) + 3);
    sprintf(imageFile, "%s%s/%s/%s", appPath, themeDir, theme, file);

    image = RLoadImage(ctx, imageFile, 0);

    if (!image) {
        printError("Could not load image %s (%s)", imageFile, RMessageForError(RErrorCode));
    }

    free(imageFile);

    return image;
}

static Pixmap* loadShapeMask(Display *dpy, const char* appPath, const char* theme, const char* file) {    
    Pixmap	 *xpm;
    char*	 themeDir;
    char*	 imageFile;
    unsigned int width, height;
    int 	 x_hotspot, y_hotspot;
    int		 ret;

    themeDir = THEME_DIR;
    imageFile = alloc(strlen(appPath) + strlen(themeDir) + strlen(theme) + strlen(file) + 3);
    sprintf(imageFile, "%s%s/%s/%s", appPath, themeDir, theme, file);

    xpm = (Pixmap*) malloc(sizeof(Pixmap));
    ret = XReadBitmapFile(dpy, DefaultRootWindow(dpy), imageFile,
    			  &width, &height,
			  xpm,
			  &x_hotspot, &y_hotspot);

    if (ret != BitmapSuccess) {
        printError("Could not load bitmap file %s ", imageFile);
	free(xpm);
	xpm = NULL;
    }

    free(imageFile);

    return xpm;
}

Resources* allocResources(Display* dpy, const char* appPath, Preferences* preferences) {
    Resources* resources;

    resources = alloc(sizeof(Resources));

    /* Load shape mask */
    if(strcmp( preferences->panelImageMask, DEFAULT_PANELIMAGEMASK)) {
       resources->panelShapeMask = loadShapeMask(dpy, appPath, preferences->theme, preferences->panelImageMask);
       if (resources->panelShapeMask == NULL) {
           return 0;
       }
    }
    if(strcmp( preferences->powerButtonImageMask, DEFAULT_POWERBUTTONIMAGEMASK)) {
       resources->powerButtonShapeMask = loadShapeMask(dpy, appPath, preferences->theme, preferences->powerButtonImageMask);
       if (resources->powerButtonShapeMask == NULL) {
           return 0;
       }
    }
    if(strcmp( preferences->restartButtonImageMask, DEFAULT_RESTARTBUTTONIMAGEMASK)) {
       resources->restartButtonShapeMask = loadShapeMask(dpy, appPath, preferences->theme, preferences->restartButtonImageMask);
       if (resources->restartButtonShapeMask == NULL) {
           return 0;
       }
    }
    
    /* Load images */
    resources->panelImage = loadImage(appPath, preferences->theme, preferences->panelImage);
    resources->fieldLabelsImage = loadImage(appPath, preferences->theme, preferences->fieldLabelsImage);
    resources->powerButtonUpImage = loadImage(appPath, preferences->theme, preferences->powerButtonUpImage);
    resources->powerButtonDownImage = loadImage(appPath, preferences->theme, preferences->powerButtonDownImage);
    resources->restartButtonUpImage = loadImage(appPath, preferences->theme, preferences->restartButtonUpImage);
    resources->restartButtonDownImage = loadImage(appPath, preferences->theme, preferences->restartButtonDownImage);
    resources->alertIcon = loadImage(appPath, preferences->theme, preferences->alertIcon);
    resources->inputUserImage = loadImage(appPath, preferences->theme, preferences->inputUserImage);
    resources->inputPasswordImage = loadImage(appPath, preferences->theme, preferences->inputPasswordImage);

    if (!resources->panelImage || !resources->fieldLabelsImage || 
	!resources->powerButtonUpImage || !resources->powerButtonDownImage || 
	!resources->restartButtonUpImage || !resources->restartButtonDownImage ||
	!resources->alertIcon || !resources->inputUserImage || !resources->inputPasswordImage) {
	return 0;
    }
	
    /* Allocate colors */
    resources->backgroundColor = getColorByName(preferences->backgroundColor);
    resources->messageColor = getColorByName(preferences->messageColor);
    resources->inputFontColor = getColorByName(preferences->inputFontColor);
    if (preferences->disableHostname == NO) {
	resources->hostnameColor = getColorByName(preferences->hostnameColor);
    }

    /* Load fonts */
    resources->messageFont = XLoadQueryFont(dpy, preferences->messageFont);
    resources->inputFont = XLoadQueryFont(dpy, preferences->inputFont);
    if (preferences->disableHostname == NO) {
	resources->hostnameFont = XLoadQueryFont(dpy, preferences->hostnameFont);
    }
    
    /* Get hostname */
    if (preferences->disableHostname == NO) {
 	resources->hostname = alloc(256);
	if (gethostname(resources->hostname, 255) != 0) {
	    preferences->disableHostname = YES;
	    printError("could not get hostname");
	}
    }

    return resources;
}
